import sys 
import getpass
s = '/home/'+getpass.getuser()
#s = '/Users/'+getpass.getuser()
sys.path.insert(1, '%s/dptqm/Structural codes/Liouvillian block diagonalization'%s)
import block_diagonalize

import numpy as np
from qutip import * 
import matplotlib.pyplot as plt

from scipy.linalg.decomp import eigvals_banded
import scipy.sparse as spr
import pickle


class Parameters():
    def __init__(self, delta, U=0, G=0, F=0, gamma=0, eta=0, kappa_phi=0, get_scGc=False):
        self.η  = eta
        self.Δ  = delta
        self.U  = U
        self.F  = F
        self.γ  = gamma
        self.κϕ = kappa_phi
        
        self.G = G
        if get_scGc and (self.G!=0 or self.η!= 0):
            self.Gc = (2*self.η*self.Δ + self.γ*self.U)/(2*np.sqrt(self.U**2 + self.η**2))        
            if self.Gc == 0: self.Gc = 1
                
            self.G *= self.Gc
            A = (self.Δ*self.U - self.η*self.γ)/(self.U**2 + self.η**2)
            B = (self.γ*self.U + 2*self.η*self.Δ)/(2*(self.U**2 + self.η**2))
            θ = lambda g: np.heaviside(g-1,1)
            self.n_sc = lambda g: (A + B*np.sqrt((g**2-1)*θ(g)))*θ(g)

class Kerr():
    def __init__(self, Ncut, p, fG=None, fdel=None, args={}):
        self.N = Ncut
        self.a = destroy(Ncut)
        self.parity = (1.j*np.pi*self.a.dag()*self.a).expm()
        
        self.p = p
        self.c_ops = []
        if self.p.η  : self.c_ops += [np.sqrt(self.p.η)*self.a**2]
        if self.p.γ  : self.c_ops += [np.sqrt(self.p.γ)*self.a]
        if self.p.κϕ : self.c_ops += [np.sqrt(self.p.κϕ)*self.a.dag()*self.a]
        
        HΔ = -self.p.Δ * self.a.dag()*self.a 
        HG = p.G/2 * (self.a.dag()**2 + self.a**2)
        
        H=0*self.a
        if self.p.U: H += self.p.U/2 * self.a.dag()**2*self.a**2
        if self.p.F: H += self.p.F * (self.a.dag() + self.a)
            
        if (fG==None) and (fdel==None):
            self.is_timedep = False
            self.H = H 
            if self.p.G: self.H += HG
            if self.p.Δ : self.H += HΔ
            self.LL = liouvillian(self.H, self.c_ops)
        else:
            self.H0 = H
            if (fG!=None):
                if self.p.Δ : self.H0 += HΔ
                self.H1_coeff, self.args = fG, args
                self.H1 = HG
                self.H  = [self.H0, [self.H1, self.H1_coeff]]
            if (fdel!=None):
                if self.p.G : self.H0 += HG
                self.H1_coeff, self.args = fdel, args
                self.H1 = HΔ
                self.H  = [self.H0, [self.H1, self.H1_coeff]]
            self.is_timedep = True        
 
    def bd(self):
        P, block_bfs, block_sizes = block_diagonalize.PermMat(self.LL)
        self.bd_L = np.dot(P, np.dot(self.LL.data, np.transpose(P)))
        num_blocks, blocks_list, bl_indices = block_diagonalize.get_blocks(self.bd_L)
        
        if num_blocks!=None:
            self.P = Qobj(P, dims = [[[self.N], [self.N]]]*2)
            self.block_bfs, self.block_sizes = block_bfs, block_sizes
            self.num_blocks, self.blocks_list, self.bl_indices = num_blocks, blocks_list, bl_indices
        
            self.m = int(self.block_sizes[0])
            self.msqrt = int(np.sqrt(self.m))
            supopdims  = [[[self.msqrt], [self.msqrt]]]*2

            if self.num_blocks == 4:
                self.iee = range(bl_indices[0],bl_indices[1])
                self.ioe = range(bl_indices[1],bl_indices[2])
                self.ieo = range(bl_indices[2],bl_indices[3])
                self.ioo = range(bl_indices[3],self.bd_L.shape[0])

                self.Lee = Qobj(blocks_list[0], dims = supopdims)
                self.Loe = Qobj(blocks_list[1], dims = supopdims)
                self.Leo = Qobj(blocks_list[2], dims = supopdims) 
                self.Loo = Qobj(blocks_list[3], dims = supopdims) 
                self.symm  = {'ee':self.iee, 'oo':self.ioo, 'oe':self.ioe, 'eo':self.ieo}
                self.Lsymm = {'ee':self.Lee, 'oo':self.Loo, 'oe':self.Loe, 'eo':self.Leo}

            elif self.num_blocks == 2:
                self.ie = range(bl_indices[0],bl_indices[1])
                self.io = range(bl_indices[1],self.bd_L.shape[0])

                self.Le = Qobj(blocks_list[0], dims = supopdims)
                self.Lo = Qobj(blocks_list[1], dims = supopdims)
                self.symm = {'e':self.ie, 'o':self.io}
            else:
                print("Unexpected number of blocks")
                return 0
        else:
            print("No blocks found")
            return 0
        
    def complete_vec(self, vec, symm, invert=False):
        vec2 = spr.dok_matrix((self.LL.shape[0], 1), dtype=np.complex128)
        i = self.symm[symm]
        vec2[i] = vec.data if vec.type=='operator-ket' else operator_to_vector(vec).data

        vec2 = Qobj(vec2,dims=[[[self.N], [self.N]], [1]], type='operator-ket')
        return (self.P.dag()*vec2 if invert else vec2)
    
    def full_completion(self, vec, symm):
        rho = Qobj(vec, dims=[[[self.msqrt],[self.msqrt]],[1]])
        rho = self.complete_vec(rho,symm,invert=True)
        rho = vector_to_operator(rho)
        #rho+=rho.dag()
        if rho.tr()!=0.:
            rho/=rho.tr()
        return rho
    

                
    def get_ss(self, method='full', get_left=False, select=1, sigma=1e-2):
        if get_left: Jlist = []
        Ψlist = []
        
        #Loop over symmerty sectors
        for i in range(int(len(self.blocks_list))):
            
            eigvs = []
            #Get right eigenvecrtor (in restricted symmetry subspace)
            block = self.blocks_list[i]
            
            if method=='full':
                _, v  = Qobj(block).eigenstates()
                eigvs.append( v[-1] )
            elif method=='arnoldi':
                e,v = spr.linalg.eigs(block, which="LM", sigma=sigma, k=select)
                arg = np.argmin(np.abs(np.real(e)))
                v   = v.T[arg]
                eigvs.append(Qobj(v))
            else:
                print("\nIncorrect method selection\n")
                break
            
            #Get left eigenvecrtor (in restricted symmetry subspace)
            if get_left:
                if method=='full':
                    R = spr.hstack(v).tocsc()
                    L = spr.linalg.inv(R)
                    eigvs.append(Qobj(L[-1].H))
                else:
                    e,v = spr.linalg.eigs(block.H, which="LM", sigma=sigma, k=select)
                    arg = np.argmin(np.abs(np.real(e)))
                    v   = v.T[arg]
                    eigvs.append(Qobj(v))
            
            #Complete vectors to full Hilbert Space
            for j in range(len(eigvs)):
                vv = spr.dok_matrix((self.LL.shape[0], 1), dtype='complex')
                vv[self.bl_indices[i]:self.bl_indices[i]+self.block_sizes[i]] = eigvs[j].data
                vv = vv.tocsr()
                vv = self.P.data.transpose().dot( vv )
                vv = vv.reshape((self.N,self.N), order='F')
                eigvs[j] = vv
            
            #Enforce proper normalization 
            Ψ = Qobj(eigvs[0]) 
            Ψ/= np.sqrt((Ψ.dag()*Ψ).tr())
            #Enforce proper phase
            if method=='arnoldi':
                s0 = Ψ.tr() if np.abs(Ψ.tr())>1e-3 else Ψ.data.A[Ψ.data.A!=0][0]
                s0/= np.abs(s0)
                Ψ/= s0
            Ψlist.append(Ψ)           
            
            if get_left:
                J  = Qobj(eigvs[1])
                if method=='arnoldi':
                    s0 = (J.dag()*Ψ).tr().conjugate()
                    J /= s0
                Jlist.append(J)
                
        if get_left:return Ψlist, Jlist
        else: return Ψlist
        
        
    
    def get_spectrum(self,ret_ss=False):
        done=False
        for i in range(int(len(self.blocks_list))):
            block = self.blocks_list[i]
            evals, evecs = Qobj(block).eigenstates()    

            ss_block_form = evecs[-1]
            evec2 = spr.dok_matrix((self.LL.shape[0], 1), dtype='complex')
            evec2[self.bl_indices[i]:self.bl_indices[i]+self.blocks_list[i].shape[0]] = ss_block_form.data
            evec2 = evec2.tocsr()
            evec2 = self.P.data.transpose().dot(evec2)
            ss= Qobj(evec2, dims=[self.LL.dims[0], [1]])
            ss=vector_to_operator(ss)

            if np.real(ss.tr())>0.05:
                ss=ss+ss.dag()
                ss/=(ss.tr())
                if expect(self.parity, ss)>0.5:
                    if ret_ss: sse=ss
                    num_even = expect(ss, self.a.dag()*self.a)
                    gap_even = evals[-2]
                else:
                    num_odd = expect(ss, self.a.dag()*self.a)
                    gap_odd = evals[-2]
                    if ret_ss: sso=ss
            elif done==False:
                gap_even_odd = np.real(evals[-1])+1.j*np.abs(np.imag(evals[-1]))
                gap_odd_even = np.real(evals[-1])-1.j*np.abs(np.imag(evals[-1]))
                if ret_ss: ssoe=ss       
                done=True
            elif ret_ss: sseo=ss 
        if ret_ss: return sse, sso, ssoe, sseo, num_even, num_odd, gap_even, gap_odd, gap_even_odd, gap_odd_even
        else: return num_even, num_odd, gap_even, gap_odd, gap_even_odd, gap_odd_even

        